/*
	Unix SMB/CIFS wrapper implementation.
	Copyright (C) 2012 by Quest Software <legal@quest.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation under version 3 of the License.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/*

DO NOT MODIFY THIS CODE TO INCLUDE ANY PROPRIETARY COMMUNICATION PROTOCOL

ALL COMMANDS SHOULD REMAIN SIMPLE STRINGS AS DESCRIBED IN cifsd.h AND OPEN
SOURCED ACCORDINGLY WITH GPLv3 LICENCE TEXT

*/

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _GNU_SOURCE
#define _FILE_OFFSET_BITS     64

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>
#include <inttypes.h>
#include <assert.h>
#include <dirent.h>
#include <limits.h>
#include <dlfcn.h>
#include <libgen.h>
#include <netdb.h>
#include <ctype.h>
#include <fnmatch.h>
#include <regex.h>
#include <iconv.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <sys/uio.h>
#include <sys/statvfs.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <zlib.h>

#include "cifsd.h"
#include <libsmbclient.h>

#define DEBUG_LEVEL 0
//#define DEBUG_LEVEL 2

#define ERROR(err,a...) \
	{ \
		fprintf (stderr, "ERROR FATAL" err "\n", ## a); \
		exit (1); \
	}   

#define REPORT_ERROR(err,a...) \
	{ \
		fprintf (fp, "ERROR " err "\n", ## a); \
		return; \
	}

#define REPORT_ERROR_NF(err,a...) \
        { \
                fprintf (fp, "ERROR_NF " err "\n", ## a); \
                return; \
        }

#define REPORT_ERRNO       REPORT_ERROR ("%d", errno)

#define REPORT_NF_ERRNO    REPORT_ERROR_NF ("%d", errno)

#define PARSE_FD(str) \
	{ \
		fd = (int) strtol (str, NULL, 10); \
		if (fd < SMBC_BASE_FD) \
			REPORT_ERROR ("Bad file descriptor"); \
	}

#define PARSE_LEN(str) \
	len = (uint64_t) strtoull (str, NULL, 10)

#ifndef __APPLE__
/*
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
static size_t strlcpy(dst, src, siz)
	char *dst;
	const char *src;
	size_t siz;
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	if (n)
		n--;		/* don't count the NUL */
	while (*s) {
		if (n) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	if (siz)
		*d = '\0';

	return(s - src);	/* count does not include NUL */
}
#endif


static void *xmalloc (size_t size)
{
	void *p;
	if (!(p = malloc (size)))
	{
		fprintf (stderr, "FATAL oom out of memory");
		exit (1);
	}
	return p;
}

static char *xstrdup (const char *str)
{
	size_t len;
	char *cp;

	len = strlen (str) + 1;
	cp = xmalloc (len);
	strncpy (cp, str, len);
	return cp;
}

static ssize_t readall (int fd, void *buf, size_t sz)
{
    int res;
    ssize_t left, total;

    if (fd < 0)
        return -1;
    left = sz, total = 0, res = -1;
    while (left && res)
    {
        if ((res = read (fd, buf, left)) < 0)
                return -1;
        buf = ((char *) buf) + res;
        left -= res, total += res;
    }

    if (total != sz)
        return -1;
    return total;
}

static ssize_t writeall (int fd, const void *buf, size_t sz)
{
	int res;
	ssize_t left, blk;

	if (fd < 0)
		return sz;
	left = sz;
	while (left)
	{
		blk = left;
		if ((res = write (fd, buf, blk)) < 0)
			return -1;
		buf = ((const char *) buf) + res;
		left -= res;
	}
	return sz;
}

static char *encode_path (const char *path)
{
	const char *p, *delim = "@";
	char *dest_path, *d;
	int found_delim = 0;

	dest_path = xmalloc (strlen (path) * 3 + 1);
	p = path;
	d = dest_path;

	for (; *p != '\0'; p++)
	{
		if (*p == *delim)
			found_delim = 1;

		if (found_delim && *p == '%')
		{
			*d++ = '%';
			*d++ = '2';
			*d++ = '5';
		}
		else
			*d++ = *p;
	}

	*d++ = '\0';

	return dest_path;
}

static void handle_open (const char *path, FILE *fp, int oflag)
{
	int fd;
	char *enc_path;

	if (!*path)
		REPORT_ERROR ("no file");
	fd = 0;

	enc_path = encode_path (path);

	if ((fd = smbc_open (enc_path, oflag, 0755)) < 0)
	{
		free (enc_path);
		REPORT_ERRNO;
	}

	free (enc_path);
	fprintf (fp, "OK %d\n", fd);
}

static void handle_close (const char *str, FILE *fp)
{
	int fd;

	PARSE_FD (str);
	if (smbc_close (fd))
		REPORT_ERRNO;
	fprintf (fp, "OK\n");
}

#define READ_BLK   65536

static void handle_read (const char *params, FILE *fp)
{
	const char *sep = " ";
	char *copy, *ctx, *fdstr, *lenstr, *buf;
	int fd;
	off_t len, ret, left, total, blk;

	buf = NULL;
	copy = xstrdup (params);
	if (!(fdstr = strtok_r (copy, sep, &ctx)))
		goto error;
	if (!(lenstr = strtok_r (NULL, sep, &ctx)))
		goto error;
	PARSE_FD (fdstr);
	PARSE_LEN (lenstr);

	buf = xmalloc (len);
	left = len, total = 0, ret = -1;
	while (left && ret)
	{
		blk = (left > READ_BLK) ? READ_BLK : left;
		if ((ret = smbc_read (fd, buf + total, blk)) < 0)
			goto error;
		left -= ret, total += ret;
	}
	fprintf (fp, "OK %" PRIu64 "\n", total);
	if (writeall (fileno (fp), buf, total) < 0)
		pthread_exit (NULL);
	goto done;

	error:
	fprintf (fp, "ERROR arguments\n");

	done:
	free (copy);
	if (buf)
		free (buf);
}

static void handle_write (const char *params, FILE *fp)
{
	const char *sep = " ";
	char *copy, *ctx, *fdstr, *lenstr, *buf;
	int fd;
	off_t len, ret, total;

	buf = NULL;
	copy = xstrdup (params);
	if (!(fdstr = strtok_r (copy, sep, &ctx)))
		goto error;
	if (!(lenstr = strtok_r (NULL, sep, &ctx)))
		goto error;
	PARSE_FD (fdstr);
	PARSE_LEN (lenstr);

	buf = xmalloc (len);
	if (readall (fileno (fp), buf, len) < 0)
		pthread_exit (NULL);

	total = 0;
	while (len - total)
	{
		if ((ret = smbc_write (fd, buf + total, len - total)) < 0)
		{
			fprintf (fp, "ERROR %d\n", errno);
			goto done;
		}
		total += ret;
	}
	fprintf (fp, "OK %" PRIu64 "\n", total);
	goto done;

	error:
	fprintf (fp, "ERROR arguments\n");

	done:
	free (copy);
	if (buf)
		free (buf);
}

static void handle_lseek (const char *params, FILE *fp)
{
	const char *sep = " ";
	char *copy, *ctx, *fdstr, *offstr, *whencestr;
	int fd, whence;
	off_t offset, ret;

	copy = xstrdup (params);
	if (!(fdstr = strtok_r (copy, sep, &ctx)))
		goto error;
	if (!(offstr = strtok_r (NULL, sep, &ctx)))
		goto error;
	if (!(whencestr = strtok_r (NULL, sep, &ctx)))
		goto error;
	PARSE_FD (fdstr);
	offset = strtoull (offstr, NULL, 10);
	if (strlen (whencestr) != 1)
		goto error;
	switch (*whencestr)
	{
		case 's':
			whence = SEEK_SET;
			break;
		case 'c':
			whence = SEEK_CUR;
			break;
		case 'e':
			whence = SEEK_END;
			break;
		default:
			goto error;
	}
	if ((ret = smbc_lseek (fd, offset, whence)) < 0)
		goto error;
	fprintf (fp, "OK %" PRIu64 "\n", ret);
	goto done;

	error:
	fprintf (fp, "ERROR arguments\n");

	done:
	free (copy);
}

static void handle_size (const char *path, FILE *fp)
{
	struct stat buf;
	char *enc_path;

	if (!*path)
		REPORT_ERROR ("no file");

	enc_path = encode_path (path);

	if (smbc_stat (enc_path, &buf))
	{
		free (enc_path);
		REPORT_ERRNO;
	}

	free (enc_path);
	fprintf (fp, "OK %" PRId64 "\n", buf.st_size);
}

static void handle_unlink (const char *path, FILE *fp)
{
	char *enc_path;

	if (!*path)
		REPORT_ERROR ("no file");

	enc_path = encode_path (path);

	if (smbc_unlink (enc_path))
	{
		free (enc_path);
		REPORT_NF_ERRNO;
	}

	free (enc_path);
	fprintf (fp, "OK\n");
}

static void handle_rename (const char *params, FILE *fp)
{
	char *a, *b, *end, *enc_a, *enc_b;
	int buflen, alen, blen;

	if (!*params)
		REPORT_ERROR ("no file");

	// these are more than we need
	buflen = strlen (params);
	a = xmalloc (buflen);
	b = xmalloc (buflen);

	// extract filename lengths and filenames
	alen = strtoul (params, &end, 10);
	if ((params == end) || (*end != ' '))
		REPORT_ERROR ("bad params");
	params = end + 1;
	blen = strtoul (params, &end, 10);
	if ((params == end) || (*end != ' '))
		REPORT_ERROR ("bad params");
	params = end + 1;
	if (strlen (params) != (alen + blen + 1))
		REPORT_ERROR ("bad params");
	memset (a, 0, buflen);
	memset (b, 0, buflen);
	strncpy (a, params, alen);
	strncpy (b, params + alen + 1, blen);

	enc_a = encode_path (a);
	free (a);
	enc_b = encode_path (b);
	free (b);

	// do the rename
	if (smbc_rename (enc_a, enc_b))
	{
		free (enc_a);
		free (enc_b);
		REPORT_ERRNO;
	}

	free (enc_a);
	free (enc_b);
	fprintf (fp, "OK\n");
}

static void handle_opendir (const char *path, FILE *fp)
{
	int fd;
	char *enc_path;

	if (!*path)
		REPORT_ERROR ("no directory");
	fd = 0;

	enc_path = encode_path (path);

	if ((fd = smbc_opendir (enc_path)) < 0)
	{
		free (enc_path);
		REPORT_NF_ERRNO;
	}

	free (enc_path);
	fprintf (fp, "OK %d\n", fd);
}

static void handle_readdir (const char *params, FILE *fp)
{
	int fd;
	struct smbc_dirent *entry;

	if (!*params)
		REPORT_ERROR ("no directory");
	PARSE_FD (params);
	if (!(entry = smbc_readdir (fd)))
		fprintf (fp, "OK\n"); // done
	else
	{
		fprintf (fp, "OK %s\n", entry->name);
	}
}

static void handle_closedir (const char *params, FILE *fp)
{
	int fd;

	if (!*params)
		REPORT_ERROR ("no directory");
	PARSE_FD (params);
	smbc_closedir (fd);
	fprintf (fp, "OK\n");
}

static void handle_mkdir (const char *path, FILE *fp)
{
	char *enc_path;

	if (!*path)
		REPORT_ERROR ("no directory");

	enc_path = encode_path (path);

	if (smbc_mkdir (enc_path, 0700))
	{
		free (enc_path);
		REPORT_ERRNO;
	}

	free (enc_path);
	fprintf (fp, "OK\n");
}

static void handle_rmdir (const char *path, FILE *fp)
{
	char *enc_path;

	if (!*path)
		REPORT_ERROR ("no directory");

	enc_path = encode_path (path);

	if (smbc_rmdir (enc_path))
	{
		free (enc_path);
		REPORT_ERRNO;
	}

	free (enc_path);
	fprintf (fp, "OK\n");
}

static void handle_access (const char *path, FILE *fp)
{
	struct stat buf;
	char *enc_path;

	if (!*path)
		REPORT_ERROR ("no path");

	enc_path = encode_path (path);

	if (smbc_stat (enc_path, &buf))
	{
		free (enc_path);
		REPORT_ERRNO;
	}

	free (enc_path);
	fprintf (fp, "OK\n");
}

static void client_handler (int sock)
{
	char *buf, *params;
	int debug, len;
	FILE *fp;

	fp = fdopen (sock, "r+");
	setvbuf (fp, NULL, _IONBF, 0);
	debug = getenv ("CIFSD_DEBUG") != NULL;
	buf = xmalloc (len = 1024 * 1024);

	while (fgets (buf, len, fp))
	{
		if (debug)
			fprintf (stderr, ">>> %s", buf);

		// strip newline
		buf[strlen (buf) - 1] = 0;

		// quit is easy
		if (!strcasecmp (buf, CIFS_CMD_QUIT))
			return;

		// the rest have params
		params = strstr (buf, " ") + 1;

		if (!strncasecmp (buf, CIFS_CMD_OPEN_R " ", 7))
			handle_open (params, fp, O_RDONLY);
		else if (!strncasecmp (buf, CIFS_CMD_OPEN_W " ", 7))
			handle_open (params, fp, O_WRONLY | O_CREAT);
		else if (!strncasecmp (buf, CIFS_CMD_CLOSE " ", 6))
			handle_close (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_READ " ", 5))
			handle_read (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_WRITE " ", 6))
			handle_write (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_LSEEK " ", 6))
			handle_lseek (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_SIZE " ", 5))
			handle_size (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_UNLINK " ", 7))
			handle_unlink (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_RENAME " ", 7))
			handle_rename (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_OPENDIR " ", 8))
			handle_opendir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_READDIR " ", 8))
			handle_readdir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_CLOSEDIR " ", 9))
			handle_closedir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_MKDIR " ", 6))
			handle_mkdir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_RMDIR " ", 6))
			handle_rmdir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_ACCESS " ", 7))
			handle_access (params, fp);
		else
			fprintf (fp, "ERROR\n");
	}
	fclose (fp);
}

static void get_auth_data (const char *srv, const char *shr, char *wg,
	int wglen, char *un, int unlen, char *pw, int pwlen)
{
	char *str;

	if ((str = getenv ("CIFS_DOMAIN")))
		strlcpy (wg, str, wglen);
	if ((str = getenv ("CIFS_USERNAME")))
		strlcpy (un, str, unlen);
	if ((str = getenv ("CIFS_PASSWORD")))
		strlcpy (pw, str, pwlen);
}

int main (void)
{
	SMBCCTX *smb_ctx;
	unsigned long smb_ctx_timeout = CIFS_DEFAULT_TIMEOUT; /* milliseconds */
	char *str, *junk;

	// check for ctx timeout val being set
	if ((str = getenv ("CIFS_TIMEOUT"))) /* seconds */
	{
		smb_ctx_timeout = strtoul (str, &junk, 10);
		if (*junk)
			ERROR ("value of CIFS_TIMEOUT environment variable is invalid");
		smb_ctx_timeout *= 1000;
		if (smb_ctx_timeout <= 0 || smb_ctx_timeout > INT_MAX)
			ERROR ("value of CIFS_TIMEOUT environment variable is invalid");
	}

	// setup smb client
	if (smbc_init (get_auth_data, DEBUG_LEVEL))
		ERROR ("smbc_init");
	smb_ctx = smbc_set_context (NULL);
	smbc_setOptionSmbEncryptionLevel (smb_ctx, SMBC_ENCRYPTLEVEL_REQUEST);
	smbc_setTimeout (smb_ctx, smb_ctx_timeout);

	// handle parent process
	client_handler (STDIN_FILENO);
	return 0;
}

